"""Client to Python Discord Pixels."""
from .autodraw import AutoDrawer
from .canvas import Canvas, Pixel
from .client import Client
from .colours import Colour, Colour as Colours
from .errors import *
